/*==============================================================================
    FILE:           user_msd_host.c

    DESCRIPTION:

    Copyright (c) 2020 Closed Loop Design, LLC

    This software is supplied "AS IS" without any warranties, express, implied
    or statutory, including but not limited to the implied warranties of fitness
    for purpose, satisfactory quality and non-infringement. Closed Loop Design LLC
    extends you a royalty-free right to use, reproduce and distribute this source
    file as well as executable files created using this source file for use with
    Analog Devices SC5xx family processors only. Nothing else gives you
    the right to use this source file.

==============================================================================*/
/*!
 * @file      user_msd_host.c
 * @brief
 *
 * @details
 *
 *
 *            Copyright (c) 2020 Closed Loop Design, LLC
 *
 *            This software is supplied "AS IS" without any warranties, express, implied
 *            or statutory, including but not limited to the implied warranties of fitness
 *            for purpose, satisfactory quality and non-infringement. Closed Loop Design LLC
 *            extends you a royalty-free right to use, reproduce and distribute this source
 *            file as well as executable files created using this source file for use with
 *            Analog Devices SC5xx family processors only. Nothing else gives you
 *            the right to use this source file.
 *
 */

#include "user_msd_host.h"
#include "cld_sc5xx_msd_usb_host_lib.h"
#include <services/gpio/adi_gpio.h>
#include <services/int/adi_int.h>
#include <services/tmr/adi_tmr.h>
#include <string.h>
#include <stdio.h>

#include "sys_services_uart.h"


#define SWAP_ENDIAN_4_BYTES(x)   (  (x << 24)                |  \
                                    ((x <<  8) & 0x00ff0000)  | \
                                    ((x >>  8) & 0x0000ff00)  | \
                                     (x >> 24)                  \
                                  )

/* ARM processor interrupt request IDs. */
#define ADI_INTR_BASE                         32         /* ARM A5 GIC offset needed for interrupt indexing for ADSP-SC58x */

#define USB0_STAT_IRQn          (ADI_INTR_BASE + 124)    /*!< Status/FIFO Data Ready */
#define USB0_DATA_IRQn          (ADI_INTR_BASE + 125)    /*!< DMA Status/Transfer Complete */


static void user_msd_host_read_capasity_rx (unsigned long data_transfer_length);
static void user_msd_host_read_capasity_failed (CLD_SC5xx_MSD_Host_Cmd_Status status);
static void user_msd_host_usb_event (CLD_USB_Event event);
static void user_msd_host_usb0_isr(uint32_t Event, void *pArg);
static void user_msd_host_read_disk_successful_callback (unsigned long data_transfer_length);
static void user_msd_host_read_disk_failed_callback (CLD_SC5xx_MSD_Host_Cmd_Status status);
static void user_msd_host_write_disk_successful_callback (unsigned long data_transfer_length);
static void user_msd_host_write_disk_failed_callback (CLD_SC5xx_MSD_Host_Cmd_Status status);
static void user_msd_host_get_lun_callback_success (void);
static void user_msd_host_get_lun_failed (CLD_SC5xx_MSD_Host_Ctrl_Req_Status reason);
static void user_msd_host_bulk_reset_callback_success (void);
static void user_msd_host_bulk_reset_failed (CLD_SC5xx_MSD_Host_Ctrl_Req_Status reason);
static void user_msd_host_read_capasity_rx (unsigned long data_transfer_length);
static void user_msd_host_read_capasity_failed (CLD_SC5xx_MSD_Host_Cmd_Status status);
static void user_msd_host_lib_status (unsigned short status_code, void * p_additional_data, unsigned short additional_data_size);

/*!< CLD SC5xx MSD USB Host library initialization data. */
static CLD_SC5xx_MSD_Host_Lib_Init_Params user_msd_host_lib_params =
{
    .usb_port                                           = SC58x_USB0,
    .enable_dma                                         = CLD_TRUE,

    .use_built_in_vbus_ctrl                             = CLD_TRUE,
    .vbus_ctrl_open_drain                               = CLD_FALSE,
    .vbus_ctrl_inverted                                 = CLD_FALSE,
    .vbus_en_port                                       = CLD_GPIO_PORT_E,
    .vbus_en_pin                                        = CLD_GPIO_PIN_11,

    .fp_cld_usb_event_callback                          = user_msd_host_usb_event,

    .fp_cld_lib_status                                  = user_msd_host_lib_status,
};

typedef enum
{
    USER_MSD_HOST_CMD_IDLE = 0,
    USER_MSD_HOST_CMD_MOUNT,
    USER_MSD_HOST_CMD_READ_DIR,
    USER_MSD_HOST_CMD_UNMOUNT,
    USER_MSD_HOST_WRITE,
	USER_MSD_HOST_READ,
} User_MSD_Host_Active_Cmd;

typedef struct
{
    CLD_Boolean initialized;
    CLD_Boolean device_connected;
    CLD_Boolean msd_init;

    CLD_Boolean capasity_valid;
    CLD_MSD_Read_Capasity_Response drive_capasity;

    CLD_RV read_status;
    CLD_RV write_status;

    FATFS fatFS; /* File system object */
    FIL fileObj;

    User_MSD_Host_Active_Cmd active_ff_cmd;
    unsigned char get_capasity_retry_cnt;

} User_Msd_Host_Data;

static User_Msd_Host_Data user_msd_host_data __attribute__ ((section(".l3_uncached_data"))) =
{
    .initialized = CLD_FALSE,
    .device_connected = CLD_FALSE,
    .msd_init = CLD_FALSE,
    .capasity_valid = CLD_FALSE,
    .active_ff_cmd = USER_MSD_HOST_CMD_IDLE,
    .get_capasity_retry_cnt = 5,
};

static unsigned char max_lun;

/**
 * Initializes the CLD SC5xx MSD USB Host library.
 *
 * @retval USER_AUDIO_INIT_SUCCESS.
 * @retval USER_AUDIO_INIT_ONGOING.
 * @retval USER_AUDIO_INIT_FAILED.
 */
User_Init_Return_Code user_msd_host_init (void)
{
    static unsigned char state = 0;
    CLD_RV cld_rv = CLD_ONGOING;
    User_Init_Return_Code rv = USER_AUDIO_INIT_ONGOING;

    switch (state)
    {
        case 0:
        	/* Configure the GPIO pins connected to the SC573 EZ-Board general purpose LEDs (LED10, LED11, LED12) */
			adi_gpio_SetDirection(ADI_GPIO_PORT_E, ADI_GPIO_PIN_13, ADI_GPIO_DIRECTION_OUTPUT);
			adi_gpio_Clear(ADI_GPIO_PORT_E, ADI_GPIO_PIN_13);
			adi_gpio_SetDirection(ADI_GPIO_PORT_A, ADI_GPIO_PIN_9, ADI_GPIO_DIRECTION_OUTPUT);
			adi_gpio_Clear(ADI_GPIO_PORT_A, ADI_GPIO_PIN_9);
			adi_gpio_SetDirection(ADI_GPIO_PORT_E, ADI_GPIO_PIN_1, ADI_GPIO_DIRECTION_OUTPUT);
			adi_gpio_Clear(ADI_GPIO_PORT_E, ADI_GPIO_PIN_1);
            state++;
        break;
        case 1:
            /* Initialize the CLD Audio v2.0 with CDC Library */
            cld_rv = cld_sc5xx_msd_host_lib_init(&user_msd_host_lib_params);

            if (cld_rv == CLD_SUCCESS)
            {
                /* Register and enable the USB interrupt ISR functions */
                adi_int_InstallHandler(USB0_STAT_IRQn, user_msd_host_usb0_isr, NULL, 1);
                adi_int_InstallHandler(USB0_DATA_IRQn, user_msd_host_usb0_isr, NULL, 1);

                user_msd_host_data.initialized = CLD_TRUE;
                rv = USER_AUDIO_INIT_SUCCESS;
            }
            else if (cld_rv == CLD_FAIL)
            {
                rv = USER_AUDIO_INIT_FAILED;
            }
            else
            {
                rv = USER_AUDIO_INIT_ONGOING;
            }
        break;
    }
    return rv;
}


/**
 * Example User Mainline function.
 *
 */
void user_msd_host_main (void)
{
    static CLD_Time main_time = 0;
    static const char drive_num = 0;
    static DIR dir;        /* Directory object */
    FILINFO fno;    /* File information object */
    FRESULT rc;     /* Result code */
    static char string[256];
    unsigned int write_length;
    unsigned int read_length;
    #define WRITE_FILE_STRING "Fat file write test\n\r"

    static unsigned char test_rd[64]  __attribute__ ((section(".l3_uncached_data")));

    if (cld_time_passed_ms(main_time) >= 250u)
    {
        main_time = cld_time_get();
        adi_gpio_Toggle(ADI_GPIO_PORT_E, ADI_GPIO_PIN_13);
    }

    switch (user_msd_host_data.active_ff_cmd)
    {
        case USER_MSD_HOST_CMD_IDLE:

        break;
        case USER_MSD_HOST_CMD_MOUNT:
            rc = f_mount(&user_msd_host_data.fatFS, &drive_num, 1);     /* Register volume work area (never fails) */
            if (rc == FR_OK)
            {
                sys_services_uart_print("Drive mounted\r\n");
            }
            user_msd_host_data.active_ff_cmd = USER_MSD_HOST_CMD_IDLE;
        break;
        case USER_MSD_HOST_CMD_READ_DIR:
            rc = f_opendir(&dir, "");
            if (rc)
            {
                /* Error opening root directory */
            }
            else
            {
                sys_services_uart_print("\r\nDirectory listing...\r\n");
                while (sys_services_uart_idle() != CLD_TRUE)
                {
                    ;
                }
                for (;;)
                {
                    /* Read a directory item */
                    rc = f_readdir(&dir, &fno);
                    if (rc || !fno.fname[0])
                    {
                        break;                  /* Error or end of dir */
                    }
                    if (fno.fattrib & AM_DIR)
                    {
                        sprintf(string, "   <dir>  %s\r\n", fno.fname);
                        sys_services_uart_print(string);
                        while (sys_services_uart_idle() != CLD_TRUE)
                        {
                            ;
                        }
                    }
                    else
                    {
                        sprintf(string, "   %s\r\n", fno.fname);
                        sys_services_uart_print(string);
                        while (sys_services_uart_idle() != CLD_TRUE)
                        {
                            ;
                        }
                    }
                }
                if (rc)
                {
                    /* Error reading the directory */
                }
            }
            user_msd_host_data.active_ff_cmd = USER_MSD_HOST_CMD_IDLE;
        break;
        case USER_MSD_HOST_CMD_UNMOUNT:
            f_unmount(&drive_num);
            sys_services_uart_print("Drive unmounted\r\n");
            user_msd_host_data.active_ff_cmd = USER_MSD_HOST_CMD_IDLE;
        break;
        case USER_MSD_HOST_WRITE:
            rc = f_open(&user_msd_host_data.fileObj, "msd_test.txt", (FA_WRITE | FA_OPEN_ALWAYS));
            if (rc == 0)
            {
                rc = f_write(&user_msd_host_data.fileObj, WRITE_FILE_STRING, strlen(WRITE_FILE_STRING), &write_length);
                if (!rc && write_length == strlen(WRITE_FILE_STRING))
                {
                    rc = f_close(&user_msd_host_data.fileObj);
                    user_msd_host_data.active_ff_cmd = USER_MSD_HOST_CMD_IDLE;
                    sys_services_uart_print("File Written\r\n");
                }
            }

        break;
        case USER_MSD_HOST_READ:
        	rc = f_open(&user_msd_host_data.fileObj, "msd_test.txt", (FA_READ | FA_OPEN_ALWAYS));
			if (rc == 0)
			{
				rc = f_read(&user_msd_host_data.fileObj, test_rd, strlen(WRITE_FILE_STRING), &read_length);

				if (!rc && read_length == strlen(WRITE_FILE_STRING))
				{
				   rc = f_close(&user_msd_host_data.fileObj);
				   user_msd_host_data.active_ff_cmd = USER_MSD_HOST_CMD_IDLE;
				   sys_services_uart_print("File Read\r\n");
				}
			}
        break;
    }
}

/**
 * Function Called when a USB event occurs on the USB Audio USB Port.
 *
 * @param event - identifies which USB event has occurred.
 */
static void user_msd_host_usb_event (CLD_USB_Event event)
{
    static CLD_SC5xx_MSD_Host_Get_Max_Lun_Params get_lun_params =
    {
        .p_max_lun = &max_lun,
        .fp_cmd_successful_callback = user_msd_host_get_lun_callback_success,
        .fp_cmd_failed_callback = user_msd_host_get_lun_failed,
    };
    static CLD_SC5xx_MSD_Host_Bulk_Only_Reset_Params reset_params =
    {
        .fp_cmd_successful_callback = user_msd_host_bulk_reset_callback_success,
        .fp_cmd_failed_callback = user_msd_host_bulk_reset_failed,
    };


    switch (event)
    {
        case CLD_USB_ENUMERATED_CONFIGURED_HS:
        case CLD_USB_ENUMERATED_CONFIGURED_FS:
        	sys_services_uart_print("\n\rDrive connected\r\n");
            cld_sc5xx_msd_host_lib_get_max_lun(&get_lun_params);
            cld_sc5xx_msd_host_lib_bulk_only_reset(&reset_params);
            user_msd_host_data.device_connected = CLD_TRUE;
            user_msd_host_data.msd_init = CLD_FALSE;
        break;
        case CLD_USB_MSD_DISCONNECTED:
        	sys_services_uart_print("Drive disconnected\r\n");
            user_msd_host_data.device_connected = CLD_FALSE;
            user_msd_host_data.msd_init = CLD_FALSE;
            user_msd_host_data.get_capasity_retry_cnt = 5;
        break;
    }
}

/**
 * User defined USB 0 ISR.
 *
 * @param Event, pArg.
 */
static void user_msd_host_usb0_isr(uint32_t Event, void *pArg)
{
    cld_usb0_isr_callback();
}

/**
 * User defined Timer ISR.
 *
 * @param pCBParam, Event, pArg.
 */
void user_msd_host_timer_handler(void *pCBParam, uint32_t Event, void *pArg)
{
    switch(Event)
    {
        case ADI_TMR_EVENT_DATA_INT:
            /* Call the CLD library 125 microsecond function */
            cld_time_125us_tick();
            break;

        default:
            break;
    }

    return;
}
/**
 * User defined Timer ISR - currently set to run every millisecond.
 *
 * @param pCBParam, Event, pArg.
 */
void user_msd_host_low_priority_timer_handler(void *pCBParam, uint32_t Event, void *pArg)
{
    switch(Event)
    {
        case ADI_TMR_EVENT_DATA_INT:
            if (user_msd_host_data.initialized == CLD_TRUE)
            {
            	/* The library mainline call is being called from a timer because
            	 * FatFs expects the MSD read and write functions to return once
            	 * the requested operation is complete, which requires the library mainline
            	 * function to have runtime.  This timer is configured as the lowest priority. */
                cld_sc5xx_msd_host_lib_main();
            }
            break;

        default:
            break;
    }

    return;
}

/**
 * Function called when the CLD library reports a status.
 *
 * @param status_code, p_additional_data, additional_data_size.
 */
static void user_msd_host_lib_status (unsigned short status_code, void * p_additional_data, unsigned short additional_data_size)
{
    /* Process the CLD library status */
    char * p_str = cld_lib_status_decode(status_code, p_additional_data, additional_data_size);
    sys_services_uart_print(p_str);
}

/**
 * Function called when the system services uart receives a byte.
 *
 * @param byte.
 */
void user_msd_host_rx_term_byte (unsigned char byte)
{
    switch (byte)
    {
        case 'M':
        case 'm':
            user_msd_host_data.active_ff_cmd = USER_MSD_HOST_CMD_MOUNT;
        break;
        case 'D':
        case 'd':
            user_msd_host_data.active_ff_cmd = USER_MSD_HOST_CMD_READ_DIR;
        break;
        case 'U':
        case 'u':
            user_msd_host_data.active_ff_cmd = USER_MSD_HOST_CMD_UNMOUNT;
        break;
        case 'W':
        case 'w':
            user_msd_host_data.active_ff_cmd = USER_MSD_HOST_WRITE;
        break;
        case 'R':
		case 'r':
			user_msd_host_data.active_ff_cmd = USER_MSD_HOST_READ;
		break;
    }
}

/**
 * Function called if the get max lun request is successful.
 *
 */
static void user_msd_host_get_lun_callback_success (void)
{
}

/**
 * Function called if the get max lun request failed.
 *
 */
static void user_msd_host_get_lun_failed (CLD_SC5xx_MSD_Host_Ctrl_Req_Status reason)
{
}

/**
 * Function called if the msd bulk reset request is successful.
 *
 */
static void user_msd_host_bulk_reset_callback_success (void)
{
    static CLD_SC5xx_MSD_Host_Cmd_Params cmd_params =
    {
        .p_data_transport = (unsigned char *)&user_msd_host_data.drive_capasity,
        .fp_cmd_successful_callback = user_msd_host_read_capasity_rx,
        .fp_cmd_failed_callback = user_msd_host_read_capasity_failed,
    };

    user_msd_host_data.capasity_valid = CLD_FALSE;

    /* Get the Drive Capasity */
    cmd_params.cmd = CLD_SC5xx_MSD_HOST_COMMAND_READ_CAPASITY;
    cmd_params.cmd_params.read_capasity.logical_unit_num = 0;
    cmd_params.cmd_params.read_capasity.logical_block_address = 0;
    cmd_params.data_transport_size = 8;
    cld_sc5xx_msd_host_lib_send_command(&cmd_params);
}

/**
 * Function called if the msd bulk reset request failed.
 *
 */
static void user_msd_host_bulk_reset_failed (CLD_SC5xx_MSD_Host_Ctrl_Req_Status reason)
{
}

/**
 * Function called if the msd Get Capasity command is successful.
 *
 * @param data_transfer_length.
 */
static void user_msd_host_read_capasity_rx (unsigned long data_transfer_length)
{
    if (data_transfer_length == sizeof (user_msd_host_data.drive_capasity))
    {
        user_msd_host_data.drive_capasity.block_length = SWAP_ENDIAN_4_BYTES(user_msd_host_data.drive_capasity.block_length);
        user_msd_host_data.drive_capasity.last_logical_block_address = SWAP_ENDIAN_4_BYTES(user_msd_host_data.drive_capasity.last_logical_block_address);
        user_msd_host_data.capasity_valid = CLD_TRUE;
        user_msd_host_data.get_capasity_retry_cnt = 5;
    }
    else
    {
        /* Request capasity again */
        if (user_msd_host_data.get_capasity_retry_cnt)
        {
            user_msd_host_bulk_reset_callback_success();
            user_msd_host_data.get_capasity_retry_cnt--;
        }
    }
}

/**
 * Function called if the msd Get Capasity command failed.
 *
 * @param status.
 */
static void user_msd_host_read_capasity_failed (CLD_SC5xx_MSD_Host_Cmd_Status status)
{
    user_msd_host_data.capasity_valid = CLD_FALSE;
    /* Request capasity again */
    if (user_msd_host_data.get_capasity_retry_cnt)
    {
        user_msd_host_bulk_reset_callback_success();
        user_msd_host_data.get_capasity_retry_cnt--;
    }
}

/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/

/*   The functions below are used by FatFs to access the MSD             */

/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/


/**
 * FatFs Disk Status command.
 *
 * @param pdrv.
 * @retval disk status.
 */
/*-----------------------------------------------------------------------*/
/* Get Drive Status                                                      */
/*-----------------------------------------------------------------------*/
DSTATUS disk_status (BYTE pdrv) /* Physical drive nmuber to identify the drive */
{
    DSTATUS rv = 0;

    if (pdrv)
    {
        rv = STA_NODISK;
    }
    else
    {
        if (user_msd_host_data.device_connected == CLD_FALSE)
        {
            rv = STA_NODISK;
        }
        if (user_msd_host_data.msd_init == CLD_FALSE)
        {
            rv |= STA_NOINIT;
        }
    }

    return rv;
}

/**
 * FatFs Disk initialization command.
 *
 * @param pdrv.
 * @retval drive status.
 */
DSTATUS disk_initialize (BYTE pdrv) /* Physical drive nmuber to identify the drive */
{
    DSTATUS rv = 0;
    if (pdrv)
    {
        rv = STA_NOINIT;              /* Supports only single drive */
    }
    /*  if (Stat & STA_NODISK) return Stat; *//* No card in the socket */

    else if (user_msd_host_data.device_connected == CLD_FALSE)
    {
        rv = STA_NODISK;                    /* card is already enumerated */
    }
    else
    {
        user_msd_host_data.msd_init = CLD_TRUE;

        #if !_FS_READONLY
        //TODO init real-time clook
        #endif

        rv = 0;
    }
    return rv;
}

/**
 * FatFs Disk read command.
 *
 * @param pdrv, buff, sector, count.
 * @retval command result.
 */
DRESULT disk_read (
    BYTE pdrv,      /* Physical drive nmuber to identify the drive */
    BYTE *buff,     /* Data buffer to store read data */
    LBA_t sector,   /* Start sector in LBA */
    UINT count      /* Number of sectors to read */
)
{
    DRESULT rv;
    CLD_SC5xx_MSD_Host_Cmd_Params cmd_params =
    {
        .p_data_transport = buff,
        .fp_cmd_successful_callback = user_msd_host_read_disk_successful_callback,
        .fp_cmd_failed_callback     = user_msd_host_read_disk_failed_callback,
    };

    if (pdrv || (count == 0))
    {
        rv = RES_PARERR;
    }
    else if (user_msd_host_data.capasity_valid == CLD_FALSE)
    {
        rv = RES_NOTRDY;
    }
    else
    {
        user_msd_host_data.read_status = CLD_ONGOING;

        cmd_params.cmd = CLD_SC5xx_MSD_HOST_COMMAND_READ_10;
        cmd_params.cmd_params.read_10.logical_block_address = sector;
        cmd_params.cmd_params.read_10.logical_unit_num = 0;
        cmd_params.cmd_params.read_10.transfer_length = count;
        cmd_params.data_transport_size = count * user_msd_host_data.drive_capasity.block_length;
        cld_sc5xx_msd_host_lib_send_command(&cmd_params);

        while (user_msd_host_data.read_status == CLD_ONGOING)
        {
            ;
        }

        if (user_msd_host_data.read_status == CLD_SUCCESS)
        {
            rv = RES_OK;
        }
        else
        {
            rv = RES_ERROR;
        }
    }

    return rv;
}

/**
 * Function called if the disk_read MSD Read command is successful.
 *
 * @param data_transfer_length.
 */
static void user_msd_host_read_disk_successful_callback (unsigned long data_transfer_length)
{
    user_msd_host_data.read_status = CLD_SUCCESS;
}

/**
 * Function called if the disk_read MSD Read command failed.
 *
 * @param status.
 */
static void user_msd_host_read_disk_failed_callback (CLD_SC5xx_MSD_Host_Cmd_Status status)
{
    user_msd_host_data.read_status = CLD_FAIL;
}

#if FF_FS_READONLY == 0
/**
 * FatFs Disk write command.
 *
 * @param pdrv, buff, sector, count.
 * @retval command result.
 */
DRESULT disk_write (
    BYTE pdrv,          /* Physical drive nmuber to identify the drive */
    const BYTE *buff,   /* Data to be written */
    LBA_t sector,       /* Start sector in LBA */
    UINT count          /* Number of sectors to write */
)
{
    DRESULT rv;
    CLD_SC5xx_MSD_Host_Cmd_Params cmd_params =
    {
        .p_data_transport = (unsigned char*)buff,
        .fp_cmd_successful_callback = user_msd_host_write_disk_successful_callback,
        .fp_cmd_failed_callback     = user_msd_host_write_disk_failed_callback,
    };

    if (pdrv || (count == 0))
    {
        rv = RES_PARERR;
    }
    else if (user_msd_host_data.capasity_valid == CLD_FALSE)
    {
        rv = RES_NOTRDY;
    }
    else
    {
        user_msd_host_data.write_status = CLD_ONGOING;

        cmd_params.cmd = CLD_SC5xx_MSD_HOST_COMMAND_WRITE_10;
        cmd_params.cmd_params.read_10.logical_block_address = sector;
        cmd_params.cmd_params.read_10.logical_unit_num = 0;
        cmd_params.cmd_params.read_10.transfer_length = count;
        cmd_params.data_transport_size = count * user_msd_host_data.drive_capasity.block_length;
        cld_sc5xx_msd_host_lib_send_command(&cmd_params);

        while (user_msd_host_data.write_status == CLD_ONGOING)
        {
            ;
        }

        if (user_msd_host_data.write_status == CLD_SUCCESS)
        {
            rv = RES_OK;
        }
        else
        {
            rv = RES_ERROR;
        }
    }

    return rv;
}

#endif

/**
 * Function called if the disk_writeMSD Write command is successful.
 *
 * @param data_transfer_length.
 */
static void user_msd_host_write_disk_successful_callback (unsigned long data_transfer_length)
{
    user_msd_host_data.write_status = CLD_SUCCESS;
}

/**
 * Function called if the disk_writeMSD Write command failedl.
 *
 * @param status.
 */
static void user_msd_host_write_disk_failed_callback (CLD_SC5xx_MSD_Host_Cmd_Status status)
{
    user_msd_host_data.write_status = CLD_FAIL;
}


/**
 * FatFs io control command.
 *
 * @param pdrv, cmd, buff.
 * @retval drive status.
 */
DRESULT disk_ioctl (
    BYTE pdrv,      /* Physical drive nmuber (0..) */
    BYTE cmd,       /* Control code */
    void *buff      /* Buffer to send/receive control data */
)
{
    DRESULT rv;

    if (pdrv)
    {
        rv = RES_PARERR;
    }
    else if (user_msd_host_data.capasity_valid == CLD_FALSE)
    {
        rv = RES_NOTRDY;
    }
    else
    {
        switch (cmd)
        {
            case CTRL_SYNC: /* Make sure that no pending write process */
                rv = RES_OK;
            break;

            case GET_SECTOR_COUNT:  /* Get number of sectors on the disk (DWORD) */
                *(unsigned long *)buff = user_msd_host_data.drive_capasity.last_logical_block_address;
                rv = RES_OK;
            break;

            case GET_SECTOR_SIZE:   /* Get R/W sector size (WORD) */
                *(unsigned long *)buff = user_msd_host_data.drive_capasity.block_length;
                rv = RES_OK;
            break;

            case GET_BLOCK_SIZE:/* Get erase block size in unit of sector (DWORD) */
                /* Erase block size fixed to 4K */
                *(unsigned long *)buff = 4*1024;
                rv = RES_OK;
            break;

            default:
                rv = RES_PARERR;
            break;
        }
    }

    return rv;
}

/**
 * FatFs get time command.
 *
 * @retval current time.
 */
DWORD get_fattime (void)
{
    return 0;
}

